# This code is part of Qiskit.
#
# (C) Copyright IBM 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test cases for the pulse Schedule transforms."""
import unittest
from typing import List, Set

import numpy as np

from qiskit import pulse
from qiskit.pulse import (
    Play,
    Delay,
    Acquire,
    Schedule,
    Waveform,
    Drag,
    Gaussian,
    GaussianSquare,
    Constant,
)
from qiskit.pulse import transforms, instructions
from qiskit.pulse.channels import (
    MemorySlot,
    DriveChannel,
    AcquireChannel,
    RegisterSlot,
    SnapshotChannel,
)
from qiskit.pulse.instructions import directives
from qiskit.providers.fake_provider import FakeOpenPulse2Q
from test import QiskitTestCase  # pylint: disable=wrong-import-order
from qiskit.utils.deprecate_pulse import decorate_test_methods, ignore_pulse_deprecation_warnings


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestAlignMeasures(QiskitTestCase):
    """Test the helper function which aligns acquires."""

    @ignore_pulse_deprecation_warnings
    def setUp(self):
        super().setUp()
        with self.assertWarns(DeprecationWarning):
            self.backend = FakeOpenPulse2Q()
        self.config = self.backend.configuration()
        self.inst_map = self.backend.defaults().instruction_schedule_map
        self.short_pulse = pulse.Waveform(
            samples=np.array([0.02739068], dtype=np.complex128), name="p0"
        )

    def test_align_measures(self):
        """Test that one acquire is delayed to match the time of the later acquire."""
        sched = pulse.Schedule(name="fake_experiment")
        sched.insert(0, Play(self.short_pulse, self.config.drive(0)), inplace=True)
        sched.insert(1, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        sched.insert(10, Acquire(5, self.config.acquire(1), MemorySlot(1)), inplace=True)
        sched.insert(10, Play(self.short_pulse, self.config.measure(0)), inplace=True)
        sched.insert(11, Play(self.short_pulse, self.config.measure(0)), inplace=True)
        sched.insert(10, Play(self.short_pulse, self.config.measure(1)), inplace=True)
        aligned = transforms.align_measures([sched])[0]
        self.assertEqual(aligned.name, "fake_experiment")

        ref = pulse.Schedule(name="fake_experiment")
        ref.insert(0, Play(self.short_pulse, self.config.drive(0)), inplace=True)
        ref.insert(10, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        ref.insert(10, Acquire(5, self.config.acquire(1), MemorySlot(1)), inplace=True)
        ref.insert(19, Play(self.short_pulse, self.config.measure(0)), inplace=True)
        ref.insert(20, Play(self.short_pulse, self.config.measure(0)), inplace=True)
        ref.insert(10, Play(self.short_pulse, self.config.measure(1)), inplace=True)

        self.assertEqual(aligned, ref)

        aligned = transforms.align_measures([sched], self.inst_map, align_time=20)[0]

        ref = pulse.Schedule(name="fake_experiment")
        ref.insert(10, Play(self.short_pulse, self.config.drive(0)), inplace=True)
        ref.insert(20, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        ref.insert(20, Acquire(5, self.config.acquire(1), MemorySlot(1)), inplace=True)
        ref.insert(29, Play(self.short_pulse, self.config.measure(0)), inplace=True)
        ref.insert(30, Play(self.short_pulse, self.config.measure(0)), inplace=True)
        ref.insert(20, Play(self.short_pulse, self.config.measure(1)), inplace=True)
        self.assertEqual(aligned, ref)

    def test_align_post_u3(self):
        """Test that acquires are scheduled no sooner than the duration of the longest X gate."""
        sched = pulse.Schedule(name="fake_experiment")
        sched = sched.insert(0, Play(self.short_pulse, self.config.drive(0)))
        sched = sched.insert(1, Acquire(5, self.config.acquire(0), MemorySlot(0)))
        sched = transforms.align_measures([sched], self.inst_map)[0]
        for time, inst in sched.instructions:
            if isinstance(inst, Acquire):
                self.assertEqual(time, 4)
        sched = transforms.align_measures([sched], self.inst_map, max_calibration_duration=10)[0]
        for time, inst in sched.instructions:
            if isinstance(inst, Acquire):
                self.assertEqual(time, 10)

    def test_multi_acquire(self):
        """Test that the last acquire is aligned to if multiple acquires occur on the
        same channel."""
        sched = pulse.Schedule()
        sched.insert(0, Play(self.short_pulse, self.config.drive(0)), inplace=True)
        sched.insert(4, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        sched.insert(20, Acquire(5, self.config.acquire(1), MemorySlot(1)), inplace=True)
        sched.insert(10, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        aligned = transforms.align_measures([sched], self.inst_map)

        ref = pulse.Schedule()
        ref.insert(0, Play(self.short_pulse, self.config.drive(0)), inplace=True)
        ref.insert(20, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        ref.insert(20, Acquire(5, self.config.acquire(1), MemorySlot(1)), inplace=True)
        ref.insert(26, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        self.assertEqual(aligned[0], ref)

    def test_multiple_acquires(self):
        """Test that multiple acquires are also aligned."""
        sched = pulse.Schedule(name="fake_experiment")
        sched.insert(0, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        sched.insert(5, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        sched.insert(10, Acquire(5, self.config.acquire(1), MemorySlot(1)), inplace=True)

        ref = pulse.Schedule()
        ref.insert(10, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        ref.insert(15, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)
        ref.insert(10, Acquire(5, self.config.acquire(1), MemorySlot(1)), inplace=True)

        aligned = transforms.align_measures([sched], self.inst_map)[0]

        self.assertEqual(aligned, ref)

    def test_align_across_schedules(self):
        """Test that acquires are aligned together across multiple schedules."""
        sched1 = pulse.Schedule(name="fake_experiment")
        sched1 = sched1.insert(0, Play(self.short_pulse, self.config.drive(0)))
        sched1 = sched1.insert(10, Acquire(5, self.config.acquire(0), MemorySlot(0)))
        sched2 = pulse.Schedule(name="fake_experiment")
        sched2 = sched2.insert(3, Play(self.short_pulse, self.config.drive(0)))
        sched2 = sched2.insert(25, Acquire(5, self.config.acquire(0), MemorySlot(0)))
        schedules = transforms.align_measures([sched1, sched2], self.inst_map)
        for time, inst in schedules[0].instructions:
            if isinstance(inst, Acquire):
                self.assertEqual(time, 25)
        for time, inst in schedules[0].instructions:
            if isinstance(inst, Acquire):
                self.assertEqual(time, 25)

    def test_align_all(self):
        """Test alignment of all instructions in a schedule."""
        sched0 = pulse.Schedule()
        sched0.insert(0, Play(self.short_pulse, self.config.drive(0)), inplace=True)
        sched0.insert(10, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)

        sched1 = pulse.Schedule()
        sched1.insert(25, Play(self.short_pulse, self.config.drive(0)), inplace=True)
        sched1.insert(25, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)

        all_aligned = transforms.align_measures([sched0, sched1], self.inst_map, align_all=True)

        ref1_aligned = pulse.Schedule()
        ref1_aligned.insert(15, Play(self.short_pulse, self.config.drive(0)), inplace=True)
        ref1_aligned.insert(25, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)

        self.assertEqual(all_aligned[0], ref1_aligned)
        self.assertEqual(all_aligned[1], sched1)

        ref1_not_aligned = pulse.Schedule()
        ref1_not_aligned.insert(0, Play(self.short_pulse, self.config.drive(0)), inplace=True)
        ref1_not_aligned.insert(25, Acquire(5, self.config.acquire(0), MemorySlot(0)), inplace=True)

        all_not_aligned = transforms.align_measures(
            [sched0, sched1],
            self.inst_map,
            align_all=False,
        )
        self.assertEqual(all_not_aligned[0], ref1_not_aligned)
        self.assertEqual(all_not_aligned[1], sched1)

    def test_measurement_at_zero(self):
        """Test that acquire at t=0 works."""
        sched1 = pulse.Schedule(name="fake_experiment")
        sched1 = sched1.insert(0, Play(self.short_pulse, self.config.drive(0)))
        sched1 = sched1.insert(0, Acquire(5, self.config.acquire(0), MemorySlot(0)))
        sched2 = pulse.Schedule(name="fake_experiment")
        sched2 = sched2.insert(0, Play(self.short_pulse, self.config.drive(0)))
        sched2 = sched2.insert(0, Acquire(5, self.config.acquire(0), MemorySlot(0)))
        schedules = transforms.align_measures([sched1, sched2], max_calibration_duration=0)
        for time, inst in schedules[0].instructions:
            if isinstance(inst, Acquire):
                self.assertEqual(time, 0)
        for time, inst in schedules[0].instructions:
            if isinstance(inst, Acquire):
                self.assertEqual(time, 0)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestAddImplicitAcquires(QiskitTestCase):
    """Test the helper function which makes implicit acquires explicit."""

    @ignore_pulse_deprecation_warnings
    def setUp(self):
        super().setUp()
        with self.assertWarns(DeprecationWarning):
            self.backend = FakeOpenPulse2Q()
        self.config = self.backend.configuration()
        self.short_pulse = pulse.Waveform(
            samples=np.array([0.02739068], dtype=np.complex128), name="p0"
        )
        sched = pulse.Schedule(name="fake_experiment")
        sched = sched.insert(0, Play(self.short_pulse, self.config.drive(0)))
        sched = sched.insert(5, Acquire(5, self.config.acquire(0), MemorySlot(0)))
        sched = sched.insert(5, Acquire(5, self.config.acquire(1), MemorySlot(1)))
        self.sched = sched

    def test_add_implicit(self):
        """Test that implicit acquires are made explicit according to the meas map."""
        sched = transforms.add_implicit_acquires(self.sched, [[0, 1]])
        acquired_qubits = set()
        for _, inst in sched.instructions:
            if isinstance(inst, Acquire):
                acquired_qubits.add(inst.acquire.index)
        self.assertEqual(acquired_qubits, {0, 1})

    def test_add_across_meas_map_sublists(self):
        """Test that implicit acquires in separate meas map sublists are all added."""
        sched = transforms.add_implicit_acquires(self.sched, [[0, 2], [1, 3]])
        acquired_qubits = set()
        for _, inst in sched.instructions:
            if isinstance(inst, Acquire):
                acquired_qubits.add(inst.acquire.index)
        self.assertEqual(acquired_qubits, {0, 1, 2, 3})

    def test_dont_add_all(self):
        """Test that acquires aren't added if no qubits in the sublist aren't being acquired."""
        sched = transforms.add_implicit_acquires(self.sched, [[4, 5], [0, 2], [1, 3]])
        acquired_qubits = set()
        for _, inst in sched.instructions:
            if isinstance(inst, Acquire):
                acquired_qubits.add(inst.acquire.index)
        self.assertEqual(acquired_qubits, {0, 1, 2, 3})

    def test_multiple_acquires(self):
        """Test for multiple acquires."""
        sched = pulse.Schedule()
        acq_q0 = pulse.Acquire(1200, AcquireChannel(0), MemorySlot(0))
        sched += acq_q0
        sched += acq_q0 << sched.duration
        sched = transforms.add_implicit_acquires(sched, meas_map=[[0]])
        self.assertEqual(sched.instructions, ((0, acq_q0), (2400, acq_q0)))


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestPad(QiskitTestCase):
    """Test padding of schedule with delays."""

    def test_padding_empty_schedule(self):
        """Test padding of empty schedule."""
        self.assertEqual(pulse.Schedule(), transforms.pad(pulse.Schedule()))

    def test_padding_schedule(self):
        """Test padding schedule."""
        delay = 10
        sched = (
            Delay(delay, DriveChannel(0)).shift(10)
            + Delay(delay, DriveChannel(0)).shift(10)
            + Delay(delay, DriveChannel(1)).shift(10)
        )

        ref_sched = (
            sched  # pylint: disable=unsupported-binary-operation
            | Delay(delay, DriveChannel(0))
            | Delay(delay, DriveChannel(0)).shift(20)
            | Delay(delay, DriveChannel(1))
            | Delay(  # pylint: disable=unsupported-binary-operation
                2 * delay, DriveChannel(1)
            ).shift(20)
        )

        self.assertEqual(transforms.pad(sched), ref_sched)

    def test_padding_schedule_inverse_order(self):
        """Test padding schedule is insensitive to order in which commands were added.

        This test is the same as `test_adding_schedule` but the order by channel
        in which commands were added to the schedule to be padded has been reversed.
        """
        delay = 10
        sched = (
            Delay(delay, DriveChannel(1)).shift(10)
            + Delay(delay, DriveChannel(0)).shift(10)
            + Delay(delay, DriveChannel(0)).shift(10)
        )

        ref_sched = (
            sched  # pylint: disable=unsupported-binary-operation
            | Delay(delay, DriveChannel(0))
            | Delay(delay, DriveChannel(0)).shift(20)
            | Delay(delay, DriveChannel(1))
            | Delay(  # pylint: disable=unsupported-binary-operation
                2 * delay, DriveChannel(1)
            ).shift(20)
        )

        self.assertEqual(transforms.pad(sched), ref_sched)

    def test_padding_until_less(self):
        """Test padding until time that is less than schedule duration."""
        delay = 10

        sched = Delay(delay, DriveChannel(0)).shift(10) + Delay(delay, DriveChannel(1))

        ref_sched = sched | Delay(delay, DriveChannel(0)) | Delay(5, DriveChannel(1)).shift(10)

        self.assertEqual(transforms.pad(sched, until=15), ref_sched)

    def test_padding_until_greater(self):
        """Test padding until time that is greater than schedule duration."""
        delay = 10

        sched = Delay(delay, DriveChannel(0)).shift(10) + Delay(delay, DriveChannel(1))

        ref_sched = (
            sched  # pylint: disable=unsupported-binary-operation
            | Delay(delay, DriveChannel(0))
            | Delay(30, DriveChannel(0)).shift(20)
            | Delay(40, DriveChannel(1)).shift(10)  # pylint: disable=unsupported-binary-operation
        )

        self.assertEqual(transforms.pad(sched, until=50), ref_sched)

    def test_padding_supplied_channels(self):
        """Test padding of only specified channels."""
        delay = 10
        sched = Delay(delay, DriveChannel(0)).shift(10) + Delay(delay, DriveChannel(1))

        ref_sched = sched | Delay(delay, DriveChannel(0)) | Delay(2 * delay, DriveChannel(2))

        channels = [DriveChannel(0), DriveChannel(2)]

        self.assertEqual(transforms.pad(sched, channels=channels), ref_sched)

    def test_padding_less_than_sched_duration(self):
        """Test that the until arg is respected even for less than the input schedule duration."""
        delay = 10
        sched = Delay(delay, DriveChannel(0)) + Delay(delay, DriveChannel(0)).shift(20)
        ref_sched = sched | pulse.Delay(5, DriveChannel(0)).shift(10)
        self.assertEqual(transforms.pad(sched, until=15), ref_sched)

    def test_padding_prepended_delay(self):
        """Test that there is delay before the first instruction."""
        delay = 10
        sched = Delay(delay, DriveChannel(0)).shift(10) + Delay(delay, DriveChannel(0))

        ref_sched = (
            Delay(delay, DriveChannel(0))
            + Delay(delay, DriveChannel(0))
            + Delay(delay, DriveChannel(0))
        )

        self.assertEqual(transforms.pad(sched, until=30, inplace=True), ref_sched)

    def test_pad_no_delay_on_classical_io_channels(self):
        """Test padding does not apply to classical IO channels."""
        delay = 10
        sched = (
            Delay(delay, MemorySlot(0)).shift(20)
            + Delay(delay, RegisterSlot(0)).shift(10)
            + Delay(delay, SnapshotChannel())
        )

        ref_sched = (
            Delay(delay, MemorySlot(0)).shift(20)
            + Delay(delay, RegisterSlot(0)).shift(10)
            + Delay(delay, SnapshotChannel())
        )

        self.assertEqual(transforms.pad(sched, until=15), ref_sched)


def get_pulse_ids(schedules: List[Schedule]) -> Set[int]:
    """Returns ids of pulses used in Schedules."""
    ids = set()
    for schedule in schedules:
        for _, inst in schedule.instructions:
            ids.add(inst.pulse.id)
    return ids


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestCompressTransform(QiskitTestCase):
    """Compress function test."""

    def test_with_duplicates(self):
        """Test compression of schedule."""
        schedule = Schedule()
        drive_channel = DriveChannel(0)
        schedule += Play(Waveform([0.0, 0.1]), drive_channel)
        schedule += Play(Waveform([0.0, 0.1]), drive_channel)

        compressed_schedule = transforms.compress_pulses([schedule])
        original_pulse_ids = get_pulse_ids([schedule])
        compressed_pulse_ids = get_pulse_ids(compressed_schedule)

        self.assertEqual(len(compressed_pulse_ids), 1)
        self.assertEqual(len(original_pulse_ids), 2)
        self.assertTrue(next(iter(compressed_pulse_ids)) in original_pulse_ids)

    def test_sample_pulse_with_clipping(self):
        """Test sample pulses with clipping."""
        schedule = Schedule()
        drive_channel = DriveChannel(0)
        schedule += Play(Waveform([0.0, 1.0]), drive_channel)
        schedule += Play(Waveform([0.0, 1.001], epsilon=1e-3), drive_channel)
        schedule += Play(Waveform([0.0, 1.0000000001]), drive_channel)

        compressed_schedule = transforms.compress_pulses([schedule])
        original_pulse_ids = get_pulse_ids([schedule])
        compressed_pulse_ids = get_pulse_ids(compressed_schedule)

        self.assertEqual(len(compressed_pulse_ids), 1)
        self.assertEqual(len(original_pulse_ids), 3)
        self.assertTrue(next(iter(compressed_pulse_ids)) in original_pulse_ids)

    def test_no_duplicates(self):
        """Test with no pulse duplicates."""
        schedule = Schedule()
        drive_channel = DriveChannel(0)
        schedule += Play(Waveform([0.0, 1.0]), drive_channel)
        schedule += Play(Waveform([0.0, 0.9]), drive_channel)
        schedule += Play(Waveform([0.0, 0.3]), drive_channel)

        compressed_schedule = transforms.compress_pulses([schedule])
        original_pulse_ids = get_pulse_ids([schedule])
        compressed_pulse_ids = get_pulse_ids(compressed_schedule)
        self.assertEqual(len(original_pulse_ids), len(compressed_pulse_ids))

    def test_parametric_pulses_with_duplicates(self):
        """Test with parametric pulses."""
        schedule = Schedule()
        drive_channel = DriveChannel(0)
        schedule += Play(Gaussian(duration=25, sigma=4, amp=0.5, angle=np.pi / 2), drive_channel)
        schedule += Play(Gaussian(duration=25, sigma=4, amp=0.5, angle=np.pi / 2), drive_channel)
        schedule += Play(GaussianSquare(duration=150, amp=0.2, sigma=8, width=140), drive_channel)
        schedule += Play(GaussianSquare(duration=150, amp=0.2, sigma=8, width=140), drive_channel)
        schedule += Play(Constant(duration=150, amp=0.5, angle=0.7), drive_channel)
        schedule += Play(Constant(duration=150, amp=0.5, angle=0.7), drive_channel)
        schedule += Play(Drag(duration=25, amp=0.4, angle=-0.3, sigma=7.8, beta=4), drive_channel)
        schedule += Play(Drag(duration=25, amp=0.4, angle=-0.3, sigma=7.8, beta=4), drive_channel)

        compressed_schedule = transforms.compress_pulses([schedule])
        original_pulse_ids = get_pulse_ids([schedule])
        compressed_pulse_ids = get_pulse_ids(compressed_schedule)
        self.assertEqual(len(original_pulse_ids), 8)
        self.assertEqual(len(compressed_pulse_ids), 4)

    def test_parametric_pulses_with_no_duplicates(self):
        """Test parametric pulses with no duplicates."""
        schedule = Schedule()
        drive_channel = DriveChannel(0)
        schedule += Play(Gaussian(duration=25, sigma=4, amp=0.5, angle=np.pi / 2), drive_channel)
        schedule += Play(Gaussian(duration=25, sigma=4, amp=0.49, angle=np.pi / 2), drive_channel)
        schedule += Play(GaussianSquare(duration=150, amp=0.2, sigma=8, width=140), drive_channel)
        schedule += Play(GaussianSquare(duration=150, amp=0.19, sigma=8, width=140), drive_channel)
        schedule += Play(Constant(duration=150, amp=0.5, angle=0.3), drive_channel)
        schedule += Play(Constant(duration=150, amp=0.51, angle=0.3), drive_channel)
        schedule += Play(Drag(duration=25, amp=0.5, angle=0.5, sigma=7.8, beta=4), drive_channel)
        schedule += Play(Drag(duration=25, amp=0.5, angle=0.51, sigma=7.8, beta=4), drive_channel)

        compressed_schedule = transforms.compress_pulses([schedule])
        original_pulse_ids = get_pulse_ids([schedule])
        compressed_pulse_ids = get_pulse_ids(compressed_schedule)
        self.assertEqual(len(original_pulse_ids), len(compressed_pulse_ids))

    def test_with_different_channels(self):
        """Test with different channels."""
        schedule = Schedule()
        schedule += Play(Waveform([0.0, 0.1]), DriveChannel(0))
        schedule += Play(Waveform([0.0, 0.1]), DriveChannel(1))

        compressed_schedule = transforms.compress_pulses([schedule])
        original_pulse_ids = get_pulse_ids([schedule])
        compressed_pulse_ids = get_pulse_ids(compressed_schedule)
        self.assertEqual(len(original_pulse_ids), 2)
        self.assertEqual(len(compressed_pulse_ids), 1)

    def test_sample_pulses_with_tolerance(self):
        """Test sample pulses with tolerance."""
        schedule = Schedule()
        schedule += Play(Waveform([0.0, 0.1001], epsilon=1e-3), DriveChannel(0))
        schedule += Play(Waveform([0.0, 0.1], epsilon=1e-3), DriveChannel(1))

        compressed_schedule = transforms.compress_pulses([schedule])
        original_pulse_ids = get_pulse_ids([schedule])
        compressed_pulse_ids = get_pulse_ids(compressed_schedule)
        self.assertEqual(len(original_pulse_ids), 2)
        self.assertEqual(len(compressed_pulse_ids), 1)

    def test_multiple_schedules(self):
        """Test multiple schedules."""
        schedules = []
        for _ in range(2):
            schedule = Schedule()
            drive_channel = DriveChannel(0)
            schedule += Play(Waveform([0.0, 0.1]), drive_channel)
            schedule += Play(Waveform([0.0, 0.1]), drive_channel)
            schedule += Play(Waveform([0.0, 0.2]), drive_channel)
            schedules.append(schedule)

        compressed_schedule = transforms.compress_pulses(schedules)
        original_pulse_ids = get_pulse_ids(schedules)
        compressed_pulse_ids = get_pulse_ids(compressed_schedule)
        self.assertEqual(len(original_pulse_ids), 6)
        self.assertEqual(len(compressed_pulse_ids), 2)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestAlignSequential(QiskitTestCase):
    """Test sequential alignment transform."""

    def test_align_sequential(self):
        """Test sequential alignment without a barrier."""
        context = transforms.AlignSequential()

        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)

        schedule = pulse.Schedule()
        schedule.insert(1, instructions.Delay(3, d0), inplace=True)
        schedule.insert(4, instructions.Delay(5, d1), inplace=True)
        schedule.insert(12, instructions.Delay(7, d0), inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        # d0
        reference.insert(0, instructions.Delay(3, d0), inplace=True)
        reference.insert(8, instructions.Delay(7, d0), inplace=True)
        # d1
        reference.insert(3, instructions.Delay(5, d1), inplace=True)

        self.assertEqual(schedule, reference)

    def test_align_sequential_with_barrier(self):
        """Test sequential alignment with a barrier."""
        context = transforms.AlignSequential()

        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)

        schedule = pulse.Schedule()
        schedule.insert(1, instructions.Delay(3, d0), inplace=True)
        schedule.append(directives.RelativeBarrier(d0, d1), inplace=True)
        schedule.insert(4, instructions.Delay(5, d1), inplace=True)
        schedule.insert(12, instructions.Delay(7, d0), inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        reference.insert(0, instructions.Delay(3, d0), inplace=True)
        reference.insert(3, directives.RelativeBarrier(d0, d1), inplace=True)
        reference.insert(3, instructions.Delay(5, d1), inplace=True)
        reference.insert(8, instructions.Delay(7, d0), inplace=True)

        self.assertEqual(schedule, reference)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestAlignLeft(QiskitTestCase):
    """Test left alignment transform."""

    def test_align_left(self):
        """Test left alignment without a barrier."""
        context = transforms.AlignLeft()

        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)
        d2 = pulse.DriveChannel(2)

        schedule = pulse.Schedule()
        schedule.insert(1, instructions.Delay(3, d0), inplace=True)
        schedule.insert(17, instructions.Delay(11, d2), inplace=True)

        sched_grouped = pulse.Schedule()
        sched_grouped += instructions.Delay(5, d1)
        sched_grouped += instructions.Delay(7, d0)
        schedule.append(sched_grouped, inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        # d0
        reference.insert(0, instructions.Delay(3, d0), inplace=True)
        reference.insert(3, instructions.Delay(7, d0), inplace=True)
        # d1
        reference.insert(3, instructions.Delay(5, d1), inplace=True)
        # d2
        reference.insert(0, instructions.Delay(11, d2), inplace=True)

        self.assertEqual(schedule, reference)

    def test_align_left_with_barrier(self):
        """Test left alignment with a barrier."""
        context = transforms.AlignLeft()

        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)
        d2 = pulse.DriveChannel(2)

        schedule = pulse.Schedule()
        schedule.insert(1, instructions.Delay(3, d0), inplace=True)
        schedule.append(directives.RelativeBarrier(d0, d1, d2), inplace=True)
        schedule.insert(17, instructions.Delay(11, d2), inplace=True)

        sched_grouped = pulse.Schedule()
        sched_grouped += instructions.Delay(5, d1)
        sched_grouped += instructions.Delay(7, d0)
        schedule.append(sched_grouped, inplace=True)
        schedule = transforms.remove_directives(context.align(schedule))

        reference = pulse.Schedule()
        # d0
        reference.insert(0, instructions.Delay(3, d0), inplace=True)
        reference.insert(3, instructions.Delay(7, d0), inplace=True)
        # d1
        reference = reference.insert(3, instructions.Delay(5, d1))
        # d2
        reference = reference.insert(3, instructions.Delay(11, d2))

        self.assertEqual(schedule, reference)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestAlignRight(QiskitTestCase):
    """Test right alignment transform."""

    def test_align_right(self):
        """Test right alignment without a barrier."""
        context = transforms.AlignRight()

        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)
        d2 = pulse.DriveChannel(2)

        schedule = pulse.Schedule()
        schedule.insert(1, instructions.Delay(3, d0), inplace=True)
        schedule.insert(17, instructions.Delay(11, d2), inplace=True)

        sched_grouped = pulse.Schedule()
        sched_grouped.insert(2, instructions.Delay(5, d1), inplace=True)
        sched_grouped += instructions.Delay(7, d0)

        schedule.append(sched_grouped, inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        # d0
        reference.insert(1, instructions.Delay(3, d0), inplace=True)
        reference.insert(4, instructions.Delay(7, d0), inplace=True)
        # d1
        reference.insert(6, instructions.Delay(5, d1), inplace=True)
        # d2
        reference.insert(0, instructions.Delay(11, d2), inplace=True)
        self.assertEqual(schedule, reference)

    def test_align_right_with_barrier(self):
        """Test right alignment with a barrier."""
        context = transforms.AlignRight()

        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)
        d2 = pulse.DriveChannel(2)

        schedule = pulse.Schedule()
        schedule.insert(1, instructions.Delay(3, d0), inplace=True)
        schedule.append(directives.RelativeBarrier(d0, d1, d2), inplace=True)
        schedule.insert(17, instructions.Delay(11, d2), inplace=True)

        sched_grouped = pulse.Schedule()
        sched_grouped.insert(2, instructions.Delay(5, d1), inplace=True)
        sched_grouped += instructions.Delay(7, d0)

        schedule.append(sched_grouped, inplace=True)
        schedule = transforms.remove_directives(context.align(schedule))

        reference = pulse.Schedule()
        # d0
        reference.insert(0, instructions.Delay(3, d0), inplace=True)
        reference.insert(7, instructions.Delay(7, d0), inplace=True)
        # d1
        reference.insert(9, instructions.Delay(5, d1), inplace=True)
        # d2
        reference.insert(3, instructions.Delay(11, d2), inplace=True)

        self.assertEqual(schedule, reference)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestAlignEquispaced(QiskitTestCase):
    """Test equispaced alignment transform."""

    def test_equispaced_with_short_duration(self):
        """Test equispaced context with duration shorter than the schedule duration."""
        context = transforms.AlignEquispaced(duration=20)

        d0 = pulse.DriveChannel(0)

        schedule = pulse.Schedule()
        for _ in range(3):
            schedule.append(Delay(10, d0), inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        reference.insert(0, Delay(10, d0), inplace=True)
        reference.insert(10, Delay(10, d0), inplace=True)
        reference.insert(20, Delay(10, d0), inplace=True)

        self.assertEqual(schedule, reference)

    def test_equispaced_with_longer_duration(self):
        """Test equispaced context with duration longer than the schedule duration."""
        context = transforms.AlignEquispaced(duration=50)

        d0 = pulse.DriveChannel(0)

        schedule = pulse.Schedule()
        for _ in range(3):
            schedule.append(Delay(10, d0), inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        reference.insert(0, Delay(10, d0), inplace=True)
        reference.insert(20, Delay(10, d0), inplace=True)
        reference.insert(40, Delay(10, d0), inplace=True)

        self.assertEqual(schedule, reference)

    def test_equispaced_with_multiple_channels_short_duration(self):
        """Test equispaced context with multiple channels and duration shorter than the total
        duration."""
        context = transforms.AlignEquispaced(duration=20)

        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)

        schedule = pulse.Schedule()
        schedule.append(Delay(10, d0), inplace=True)
        schedule.append(Delay(20, d1), inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        reference.insert(0, Delay(10, d0), inplace=True)
        reference.insert(0, Delay(20, d1), inplace=True)

        self.assertEqual(schedule, reference)

    def test_equispaced_with_multiple_channels_longer_duration(self):
        """Test equispaced context with multiple channels and duration longer than the total
        duration."""
        context = transforms.AlignEquispaced(duration=30)

        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)

        schedule = pulse.Schedule()
        schedule.append(Delay(10, d0), inplace=True)
        schedule.append(Delay(20, d1), inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        reference.insert(0, Delay(10, d0), inplace=True)
        reference.insert(10, Delay(20, d1), inplace=True)

        self.assertEqual(schedule, reference)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestAlignFunc(QiskitTestCase):
    """Test callback alignment transform."""

    @staticmethod
    def _position(ind):
        """Returns 0.25, 0.5, 0.75 for ind = 1, 2, 3."""
        return ind / (3 + 1)

    def test_numerical_with_short_duration(self):
        """Test numerical alignment context with duration shorter than the schedule duration."""
        context = transforms.AlignFunc(duration=20, func=self._position)

        d0 = pulse.DriveChannel(0)

        schedule = pulse.Schedule()
        for _ in range(3):
            schedule.append(Delay(10, d0), inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        reference.insert(0, Delay(10, d0), inplace=True)
        reference.insert(10, Delay(10, d0), inplace=True)
        reference.insert(20, Delay(10, d0), inplace=True)

        self.assertEqual(schedule, reference)

    def test_numerical_with_longer_duration(self):
        """Test numerical alignment context with duration longer than the schedule duration."""
        context = transforms.AlignFunc(duration=80, func=self._position)

        d0 = pulse.DriveChannel(0)

        schedule = pulse.Schedule()
        for _ in range(3):
            schedule.append(Delay(10, d0), inplace=True)
        schedule = context.align(schedule)

        reference = pulse.Schedule()
        reference.insert(15, Delay(10, d0), inplace=True)
        reference.insert(35, Delay(10, d0), inplace=True)
        reference.insert(55, Delay(10, d0), inplace=True)

        self.assertEqual(schedule, reference)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestFlatten(QiskitTestCase):
    """Test flattening transform."""

    def test_flatten(self):
        """Test the flatten transform."""
        context_left = transforms.AlignLeft()

        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)

        schedule = pulse.Schedule()
        schedule += instructions.Delay(3, d0)

        grouped = pulse.Schedule()
        grouped += instructions.Delay(5, d1)
        grouped += instructions.Delay(7, d0)
        # include a grouped schedule
        grouped = schedule + grouped

        # flatten the schedule inline internal groups
        flattened = transforms.flatten(grouped)

        # align all the instructions to the left after flattening
        flattened = context_left.align(flattened)
        grouped = context_left.align(grouped)

        reference = pulse.Schedule()
        # d0
        reference.insert(0, instructions.Delay(3, d0), inplace=True)
        reference.insert(3, instructions.Delay(7, d0), inplace=True)
        # d1
        reference.insert(0, instructions.Delay(5, d1), inplace=True)

        self.assertEqual(flattened, reference)
        self.assertNotEqual(grouped, reference)


class _TestDirective(directives.Directive):
    """Pulse ``RelativeBarrier`` directive."""

    def __init__(self, *channels):
        """Test directive"""
        super().__init__(operands=tuple(channels))

    @property
    def channels(self):
        return self.operands


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestRemoveDirectives(QiskitTestCase):
    """Test removing of directives."""

    def test_remove_directives(self):
        """Test that all directives are removed."""
        d0 = pulse.DriveChannel(0)
        d1 = pulse.DriveChannel(1)

        schedule = pulse.Schedule()
        schedule += _TestDirective(d0, d1)
        schedule += instructions.Delay(3, d0)
        schedule += _TestDirective(d0, d1)
        schedule = transforms.remove_directives(schedule)

        reference = pulse.Schedule()
        # d0
        reference += instructions.Delay(3, d0)
        self.assertEqual(schedule, reference)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestRemoveTrivialBarriers(QiskitTestCase):
    """Test scheduling transforms."""

    def test_remove_trivial_barriers(self):
        """Test that trivial barriers are properly removed."""
        schedule = pulse.Schedule()
        schedule += directives.RelativeBarrier()
        schedule += directives.RelativeBarrier(pulse.DriveChannel(0))
        schedule += directives.RelativeBarrier(pulse.DriveChannel(0), pulse.DriveChannel(1))
        schedule = transforms.remove_trivial_barriers(schedule)

        reference = pulse.Schedule()
        reference += directives.RelativeBarrier(pulse.DriveChannel(0), pulse.DriveChannel(1))
        self.assertEqual(schedule, reference)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Blueprint circuit object."""

from __future__ import annotations
from abc import ABC, abstractmethod

from qiskit._accelerate.circuit import CircuitData
from qiskit.circuit import QuantumCircuit, QuantumRegister, ClassicalRegister
from qiskit.circuit.parametertable import ParameterView


class BlueprintCircuit(QuantumCircuit, ABC):
    """Blueprint circuit object.

    In many applications it is necessary to pass around the structure a circuit will have without
    explicitly knowing e.g. its number of qubits, or other missing information. This can be solved
    by having a circuit that knows how to construct itself, once all information is available.

    This class provides an interface for such circuits. Before internal data of the circuit is
    accessed, the ``_build`` method is called. There the configuration of the circuit is checked.
    """

    def __init__(self, *regs, name: str | None = None) -> None:
        """Create a new blueprint circuit."""
        self._is_initialized = False
        super().__init__(*regs, name=name)
        self._qregs: list[QuantumRegister] = []
        self._cregs: list[ClassicalRegister] = []
        self._qubit_indices = {}
        self._is_built = False
        self._is_initialized = True

    @abstractmethod
    def _check_configuration(self, raise_on_failure: bool = True) -> bool:
        """Check if the current configuration allows the circuit to be built.

        Args:
            raise_on_failure: If True, raise if the configuration is invalid. If False, return
                False if the configuration is invalid.

        Returns:
            True, if the configuration is valid. Otherwise, depending on the value of
            ``raise_on_failure`` an error is raised or False is returned.
        """
        raise NotImplementedError

    @abstractmethod
    def _build(self) -> None:
        """Build the circuit."""
        if self._is_built:
            return

        # check whether the configuration is valid
        self._check_configuration()
        self._is_built = True

    def _invalidate(self) -> None:
        """Invalidate the current circuit build."""
        self._data = CircuitData(self._data.qubits, self._data.clbits)
        self.global_phase = 0
        self._is_built = False

    @property
    def qregs(self):
        """A list of the quantum registers associated with the circuit."""
        return self._qregs

    @qregs.setter
    def qregs(self, qregs):
        """Set the quantum registers associated with the circuit."""
        if not self._is_initialized:
            # Workaround to ignore calls from QuantumCircuit.__init__() which
            # doesn't expect 'qregs' to be an overridden property!
            return
        self._qregs = []
        self._ancillas = []
        self._qubit_indices = {}
        self._data = CircuitData(clbits=self._data.clbits)
        self.global_phase = 0
        self._is_built = False

        self.add_register(*qregs)

    @property
    def data(self):
        """The circuit data (instructions and context).

        Returns:
            QuantumCircuitData: a list-like object containing the :class:`.CircuitInstruction`\\ s
            for each instruction.
        """
        if not self._is_built:
            self._build()
        return super().data

    def decompose(self, gates_to_decompose=None, reps=1):
        if not self._is_built:
            self._build()
        return super().decompose(gates_to_decompose, reps)

    def draw(self, *args, **kwargs):
        if not self._is_built:
            self._build()
        return super().draw(*args, **kwargs)

    @property
    def num_parameters(self) -> int:
        """The number of parameter objects in the circuit."""
        if not self._is_built:
            self._build()
        return super().num_parameters

    @property
    def parameters(self) -> ParameterView:
        """The parameters defined in the circuit.

        This attribute returns the :class:`.Parameter` objects in the circuit sorted
        alphabetically. Note that parameters instantiated with a :class:`.ParameterVector`
        are still sorted numerically.

        Examples:

            The snippet below shows that insertion order of parameters does not matter.

            .. code-block:: python

                >>> from qiskit.circuit import QuantumCircuit, Parameter
                >>> a, b, elephant = Parameter("a"), Parameter("b"), Parameter("elephant")
                >>> circuit = QuantumCircuit(1)
                >>> circuit.rx(b, 0)
                >>> circuit.rz(elephant, 0)
                >>> circuit.ry(a, 0)
                >>> circuit.parameters  # sorted alphabetically!
                ParameterView([Parameter(a), Parameter(b), Parameter(elephant)])

            Bear in mind that alphabetical sorting might be unintuitive when it comes to numbers.
            The literal "10" comes before "2" in strict alphabetical sorting.

            .. code-block:: python

                >>> from qiskit.circuit import QuantumCircuit, Parameter
                >>> angles = [Parameter("angle_1"), Parameter("angle_2"), Parameter("angle_10")]
                >>> circuit = QuantumCircuit(1)
                >>> circuit.u(*angles, 0)
                >>> circuit.draw()
                   ┌─────────────────────────────┐
                q: ┤ U(angle_1,angle_2,angle_10) ├
                   └─────────────────────────────┘
                >>> circuit.parameters
                ParameterView([Parameter(angle_1), Parameter(angle_10), Parameter(angle_2)])

            To respect numerical sorting, a :class:`.ParameterVector` can be used.

            .. code-block:: python

                >>> from qiskit.circuit import QuantumCircuit, Parameter, ParameterVector
                >>> x = ParameterVector("x", 12)
                >>> circuit = QuantumCircuit(1)
                >>> for x_i in x:
                ...     circuit.rx(x_i, 0)
                >>> circuit.parameters
                ParameterView([
                    ParameterVectorElement(x[0]), ParameterVectorElement(x[1]),
                    ParameterVectorElement(x[2]), ParameterVectorElement(x[3]),
                    ..., ParameterVectorElement(x[11])
                ])


        Returns:
            The sorted :class:`.Parameter` objects in the circuit.
        """
        if not self._is_built:
            self._build()
        return super().parameters

    def _append(self, instruction, _qargs=None, _cargs=None, *, _standard_gate=False):
        if not self._is_built:
            self._build()
        return super()._append(instruction, _qargs, _cargs, _standard_gate=_standard_gate)

    def compose(
        self,
        other,
        qubits=None,
        clbits=None,
        front=False,
        inplace=False,
        wrap=False,
        *,
        copy=True,
        var_remap=None,
        inline_captures=False,
    ):
        if not self._is_built:
            self._build()
        return super().compose(
            other,
            qubits,
            clbits,
            front,
            inplace,
            wrap,
            copy=copy,
            var_remap=var_remap,
            inline_captures=False,
        )

    def inverse(self, annotated: bool = False):
        if not self._is_built:
            self._build()
        return super().inverse(annotated=annotated)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def size(self, *args, **kwargs):
        if not self._is_built:
            self._build()
        return super().size(*args, **kwargs)

    def to_instruction(self, parameter_map=None, label=None):
        if not self._is_built:
            self._build()
        return super().to_instruction(parameter_map, label=label)

    def to_gate(self, parameter_map=None, label=None):
        if not self._is_built:
            self._build()
        return super().to_gate(parameter_map, label=label)

    def depth(self, *args, **kwargs):
        if not self._is_built:
            self._build()
        return super().depth(*args, **kwargs)

    def count_ops(self):
        if not self._is_built:
            self._build()
        return super().count_ops()

    def num_nonlocal_gates(self):
        if not self._is_built:
            self._build()
        return super().num_nonlocal_gates()

    def num_connected_components(self, unitary_only=False):
        if not self._is_built:
            self._build()
        return super().num_connected_components(unitary_only=unitary_only)

    def copy_empty_like(self, name=None, *, vars_mode="alike"):
        if not self._is_built:
            self._build()
        cpy = super().copy_empty_like(name=name, vars_mode=vars_mode)
        # The base `copy_empty_like` will typically trigger code that `BlueprintCircuit` treats as
        # an "invalidation", so we have to manually restore properties deleted by that that
        # `copy_empty_like` is supposed to propagate.
        cpy.global_phase = self.global_phase
        return cpy

    def copy(self, name=None):
        if not self._is_built:
            self._build()
        circuit_copy = super().copy(name=name)
        circuit_copy._is_built = self._is_built
        return circuit_copy



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for Sampler."""

import unittest
import numpy as np

from qiskit import QuantumCircuit
from qiskit.circuit import Parameter
from qiskit.circuit.library import RealAmplitudes, UnitaryGate
from qiskit.primitives import Sampler, SamplerResult
from qiskit.providers import JobStatus
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestSampler(QiskitTestCase):
    """Test Sampler"""

    def setUp(self):
        super().setUp()
        hadamard = QuantumCircuit(1, 1, name="Hadamard")
        hadamard.h(0)
        hadamard.measure(0, 0)
        bell = QuantumCircuit(2, name="Bell")
        bell.h(0)
        bell.cx(0, 1)
        bell.measure_all()
        self._circuit = [hadamard, bell]
        self._target = [
            {0: 0.5, 1: 0.5},
            {0: 0.5, 3: 0.5, 1: 0, 2: 0},
        ]
        self._pqc = RealAmplitudes(num_qubits=2, reps=2)
        self._pqc.measure_all()
        self._pqc2 = RealAmplitudes(num_qubits=2, reps=3)
        self._pqc2.measure_all()
        self._pqc_params = [[0.0] * 6, [1.0] * 6]
        self._pqc_target = [{0: 1}, {0: 0.0148, 1: 0.3449, 2: 0.0531, 3: 0.5872}]
        self._theta = [
            [0, 1, 1, 2, 3, 5],
            [1, 2, 3, 4, 5, 6],
            [0, 1, 2, 3, 4, 5, 6, 7],
        ]

    def _generate_circuits_target(self, indices):
        if isinstance(indices, list):
            circuits = [self._circuit[j] for j in indices]
            target = [self._target[j] for j in indices]
        else:
            raise ValueError(f"invalid index {indices}")
        return circuits, target

    def _generate_params_target(self, indices):
        if isinstance(indices, int):
            params = self._pqc_params[indices]
            target = self._pqc_target[indices]
        elif isinstance(indices, list):
            params = [self._pqc_params[j] for j in indices]
            target = [self._pqc_target[j] for j in indices]
        else:
            raise ValueError(f"invalid index {indices}")
        return params, target

    def _compare_probs(self, prob, target):
        if not isinstance(prob, list):
            prob = [prob]
        if not isinstance(target, list):
            target = [target]
        self.assertEqual(len(prob), len(target))
        for p, targ in zip(prob, target):
            for key, t_val in targ.items():
                if key in p:
                    self.assertAlmostEqual(p[key], t_val, places=1)
                else:
                    self.assertAlmostEqual(t_val, 0, places=1)

    def test_sampler_run(self):
        """Test Sampler.run()."""
        bell = self._circuit[1]
        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            job = sampler.run(circuits=[bell])
            result = job.result()
            self.assertIsInstance(result, SamplerResult)
        self._compare_probs(result.quasi_dists, self._target[1])

    def test_sample_run_multiple_circuits(self):
        """Test Sampler.run() with multiple circuits."""
        # executes three Bell circuits
        # Argument `parameters` is optional.
        bell = self._circuit[1]
        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            result = sampler.run([bell, bell, bell]).result()
        self._compare_probs(result.quasi_dists[0], self._target[1])
        self._compare_probs(result.quasi_dists[1], self._target[1])
        self._compare_probs(result.quasi_dists[2], self._target[1])

    def test_sampler_run_with_parameterized_circuits(self):
        """Test Sampler.run() with parameterized circuits."""
        # parameterized circuit

        pqc = self._pqc
        pqc2 = self._pqc2
        theta1, theta2, theta3 = self._theta

        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            result = sampler.run([pqc, pqc, pqc2], [theta1, theta2, theta3]).result()

        # result of pqc(theta1)
        prob1 = {
            "00": 0.1309248462975777,
            "01": 0.3608720796028448,
            "10": 0.09324865232050054,
            "11": 0.41495442177907715,
        }
        self.assertDictAlmostEqual(result.quasi_dists[0].binary_probabilities(), prob1)

        # result of pqc(theta2)
        prob2 = {
            "00": 0.06282290651933871,
            "01": 0.02877144385576705,
            "10": 0.606654494132085,
            "11": 0.3017511554928094,
        }
        self.assertDictAlmostEqual(result.quasi_dists[1].binary_probabilities(), prob2)

        # result of pqc2(theta3)
        prob3 = {
            "00": 0.1880263994380416,
            "01": 0.6881971261189544,
            "10": 0.09326232720582443,
            "11": 0.030514147237179892,
        }
        self.assertDictAlmostEqual(result.quasi_dists[2].binary_probabilities(), prob3)

    def test_run_1qubit(self):
        """test for 1-qubit cases"""
        qc = QuantumCircuit(1)
        qc.measure_all()
        qc2 = QuantumCircuit(1)
        qc2.x(0)
        qc2.measure_all()

        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            result = sampler.run([qc, qc2]).result()
            self.assertIsInstance(result, SamplerResult)
        self.assertEqual(len(result.quasi_dists), 2)

        for i in range(2):
            keys, values = zip(*sorted(result.quasi_dists[i].items()))
            self.assertTupleEqual(keys, (i,))
            np.testing.assert_allclose(values, [1])

    def test_run_2qubit(self):
        """test for 2-qubit cases"""
        qc0 = QuantumCircuit(2)
        qc0.measure_all()
        qc1 = QuantumCircuit(2)
        qc1.x(0)
        qc1.measure_all()
        qc2 = QuantumCircuit(2)
        qc2.x(1)
        qc2.measure_all()
        qc3 = QuantumCircuit(2)
        qc3.x([0, 1])
        qc3.measure_all()

        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            result = sampler.run([qc0, qc1, qc2, qc3]).result()
            self.assertIsInstance(result, SamplerResult)
        self.assertEqual(len(result.quasi_dists), 4)

        for i in range(4):
            keys, values = zip(*sorted(result.quasi_dists[i].items()))
            self.assertTupleEqual(keys, (i,))
            np.testing.assert_allclose(values, [1])

    def test_run_single_circuit(self):
        """Test for single circuit case."""

        with self.subTest("No parameter"):
            circuit = self._circuit[1]
            target = self._target[1]
            param_vals = [None, [], [[]], np.array([]), np.array([[]])]
            for val in param_vals:
                with self.subTest(f"{circuit.name} w/ {val}"):
                    with self.assertWarns(DeprecationWarning):
                        sampler = Sampler()
                    with self.assertWarns(DeprecationWarning):
                        result = sampler.run(circuit, val).result()
                    self._compare_probs(result.quasi_dists, target)
                    self.assertEqual(len(result.metadata), 1)

        with self.subTest("One parameter"):
            circuit = QuantumCircuit(1, 1, name="X gate")
            param = Parameter("x")
            circuit.ry(param, 0)
            circuit.measure(0, 0)
            target = [{1: 1}]
            param_vals = [
                [np.pi],
                [[np.pi]],
                np.array([np.pi]),
                np.array([[np.pi]]),
                [np.array([np.pi])],
            ]
            for val in param_vals:
                with self.subTest(f"{circuit.name} w/ {val}"):
                    with self.assertWarns(DeprecationWarning):
                        sampler = Sampler()
                    with self.assertWarns(DeprecationWarning):
                        result = sampler.run(circuit, val).result()
                    self._compare_probs(result.quasi_dists, target)
                    self.assertEqual(len(result.metadata), 1)

        with self.subTest("More than one parameter"):
            circuit = self._pqc
            target = [self._pqc_target[0]]
            param_vals = [
                self._pqc_params[0],
                [self._pqc_params[0]],
                np.array(self._pqc_params[0]),
                np.array([self._pqc_params[0]]),
                [np.array(self._pqc_params[0])],
            ]
            for val in param_vals:
                with self.subTest(f"{circuit.name} w/ {val}"):
                    with self.assertWarns(DeprecationWarning):
                        sampler = Sampler()
                    with self.assertWarns(DeprecationWarning):
                        result = sampler.run(circuit, val).result()
                    self._compare_probs(result.quasi_dists, target)
                    self.assertEqual(len(result.metadata), 1)

    def test_run_reverse_meas_order(self):
        """test for sampler with reverse measurement order"""
        x = Parameter("x")
        y = Parameter("y")

        qc = QuantumCircuit(3, 3)
        qc.rx(x, 0)
        qc.rx(y, 1)
        qc.x(2)
        qc.measure(0, 2)
        qc.measure(1, 1)
        qc.measure(2, 0)

        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            result = sampler.run([qc] * 2, [[0, 0], [np.pi / 2, 0]]).result()
            self.assertIsInstance(result, SamplerResult)
        self.assertEqual(len(result.quasi_dists), 2)

        # qc({x: 0, y: 0})
        keys, values = zip(*sorted(result.quasi_dists[0].items()))
        self.assertTupleEqual(keys, (1,))
        np.testing.assert_allclose(values, [1])

        # qc({x: pi/2, y: 0})
        keys, values = zip(*sorted(result.quasi_dists[1].items()))
        self.assertTupleEqual(keys, (1, 5))
        np.testing.assert_allclose(values, [0.5, 0.5])

    def test_run_errors(self):
        """Test for errors with run method"""
        qc1 = QuantumCircuit(1)
        qc1.measure_all()
        qc2 = RealAmplitudes(num_qubits=1, reps=1)
        qc2.measure_all()
        qc3 = QuantumCircuit(1)
        qc4 = QuantumCircuit(1, 1)
        qc5 = QuantumCircuit(1, 1)
        with qc5.for_loop(range(5)):
            qc5.h(0)

        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            with self.subTest("set parameter values to a non-parameterized circuit"):
                with self.assertRaises(ValueError):
                    _ = sampler.run([qc1], [[1e2]])
            with self.subTest("missing all parameter values for a parameterized circuit"):
                with self.assertRaises(ValueError):
                    _ = sampler.run([qc2], [[]])
            with self.subTest("missing some parameter values for a parameterized circuit"):
                with self.assertRaises(ValueError):
                    _ = sampler.run([qc2], [[1e2]])
            with self.subTest("too many parameter values for a parameterized circuit"):
                with self.assertRaises(ValueError):
                    _ = sampler.run([qc2], [[1e2]] * 100)
            with self.subTest("no classical bits"):
                with self.assertRaises(ValueError):
                    _ = sampler.run([qc3], [[]])
            with self.subTest("no measurement"):
                with self.assertRaises(ValueError):
                    _ = sampler.run([qc4], [[]])
            with self.subTest("no measurement in control flow"):
                with self.assertRaises(ValueError):
                    _ = sampler.run([qc5], [[]])

    def test_run_empty_parameter(self):
        """Test for empty parameter"""
        n = 5
        qc = QuantumCircuit(n, n - 1)
        qc.measure(range(n - 1), range(n - 1))
        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
        with self.subTest("one circuit"):
            with self.assertWarns(DeprecationWarning):
                result = sampler.run([qc], shots=1000).result()
            self.assertEqual(len(result.quasi_dists), 1)
            for q_d in result.quasi_dists:
                quasi_dist = {k: v for k, v in q_d.items() if v != 0.0}
                self.assertDictEqual(quasi_dist, {0: 1.0})
            self.assertEqual(len(result.metadata), 1)

        with self.subTest("two circuits"):
            result = sampler.run([qc, qc], shots=1000).result()
            self.assertEqual(len(result.quasi_dists), 2)
            for q_d in result.quasi_dists:
                quasi_dist = {k: v for k, v in q_d.items() if v != 0.0}
                self.assertDictEqual(quasi_dist, {0: 1.0})
            self.assertEqual(len(result.metadata), 2)

    def test_run_numpy_params(self):
        """Test for numpy array as parameter values"""
        qc = RealAmplitudes(num_qubits=2, reps=2)
        qc.measure_all()
        k = 5
        rng = np.random.default_rng(12)
        params_array = rng.random((k, qc.num_parameters))
        params_list = params_array.tolist()
        params_list_array = list(params_array)
        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            target = sampler.run([qc] * k, params_list).result()

        with self.subTest("ndarrary"):
            result = sampler.run([qc] * k, params_array).result()
            self.assertEqual(len(result.metadata), k)
            for i in range(k):
                self.assertDictEqual(result.quasi_dists[i], target.quasi_dists[i])

        with self.subTest("list of ndarray"):
            result = sampler.run([qc] * k, params_list_array).result()
            self.assertEqual(len(result.metadata), k)
            for i in range(k):
                self.assertDictEqual(result.quasi_dists[i], target.quasi_dists[i])

    def test_run_with_shots_option(self):
        """test with shots option."""
        params, target = self._generate_params_target([1])
        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            result = sampler.run(
                circuits=[self._pqc], parameter_values=params, shots=1024, seed=15
            ).result()
        self._compare_probs(result.quasi_dists, target)

    def test_run_with_shots_option_none(self):
        """test with shots=None option. Seed is ignored then."""
        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            result_42 = sampler.run(
                [self._pqc], parameter_values=[[0, 1, 1, 2, 3, 5]], shots=None, seed=42
            ).result()
            result_15 = sampler.run(
                [self._pqc], parameter_values=[[0, 1, 1, 2, 3, 5]], shots=None, seed=15
            ).result()
        self.assertDictAlmostEqual(result_42.quasi_dists, result_15.quasi_dists)

    def test_run_shots_result_size(self):
        """test with shots option to validate the result size"""
        n = 10
        shots = 100
        qc = QuantumCircuit(n)
        qc.h(range(n))
        qc.measure_all()
        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            result = sampler.run(qc, [], shots=shots, seed=42).result()
        self.assertEqual(len(result.quasi_dists), 1)
        self.assertLessEqual(len(result.quasi_dists[0]), shots)
        self.assertAlmostEqual(sum(result.quasi_dists[0].values()), 1.0)

    def test_primitive_job_status_done(self):
        """test primitive job's status"""
        bell = self._circuit[1]
        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            job = sampler.run(circuits=[bell])
            _ = job.result()
        self.assertEqual(job.status(), JobStatus.DONE)

    def test_options(self):
        """Test for options"""
        with self.subTest("init"):
            with self.assertWarns(DeprecationWarning):
                sampler = Sampler(options={"shots": 3000})
            self.assertEqual(sampler.options.get("shots"), 3000)
        with self.subTest("set_options"):
            sampler.set_options(shots=1024, seed=15)
            self.assertEqual(sampler.options.get("shots"), 1024)
            self.assertEqual(sampler.options.get("seed"), 15)
        with self.subTest("run"):
            params, target = self._generate_params_target([1])
            with self.assertWarns(DeprecationWarning):
                result = sampler.run([self._pqc], parameter_values=params).result()
            self._compare_probs(result.quasi_dists, target)
            self.assertEqual(result.quasi_dists[0].shots, 1024)

    def test_circuit_with_unitary(self):
        """Test for circuit with unitary gate."""
        gate = UnitaryGate(np.eye(2))

        circuit = QuantumCircuit(1)
        circuit.append(gate, [0])
        circuit.measure_all()

        with self.assertWarns(DeprecationWarning):
            sampler = Sampler()
            sampler_result = sampler.run([circuit]).result()
        self.assertDictAlmostEqual(sampler_result.quasi_dists[0], {0: 1, 1: 0})


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the evolved operator ansatz."""

from ddt import ddt, data
import numpy as np

from qiskit.circuit import QuantumCircuit
from qiskit.quantum_info import SparsePauliOp, Operator, Pauli

from qiskit.circuit.library import HamiltonianGate
from qiskit.circuit.library.n_local import (
    EvolvedOperatorAnsatz,
    evolved_operator_ansatz,
    hamiltonian_variational_ansatz,
)
from qiskit.synthesis.evolution import MatrixExponential
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestEvolvedOperatorAnsatz(QiskitTestCase):
    """Test the evolved operator ansatz."""

    @data(True, False)
    def test_evolved_op_ansatz(self, use_function):
        """Test the default evolution."""
        num_qubits = 3
        ops = [Pauli("Z" * num_qubits), Pauli("Y" * num_qubits), Pauli("X" * num_qubits)]

        if use_function:
            evo = evolved_operator_ansatz(ops, 2)
        else:
            evo = EvolvedOperatorAnsatz(ops, 2)

        parameters = evo.parameters

        reference = QuantumCircuit(num_qubits)
        strings = ["z" * num_qubits, "y" * num_qubits, "x" * num_qubits] * 2
        for string, time in zip(strings, parameters):
            reference.compose(evolve(string, time), inplace=True)

        if not use_function:
            evo = evo.decompose().decompose()

        self.assertEqual(evo, reference)

    @data(True, False)
    def test_custom_evolution(self, use_function):
        """Test using another evolution than the default (e.g. matrix evolution)."""
        op = SparsePauliOp(["ZIX"])
        matrix = np.array(op)
        evolution = MatrixExponential()

        if use_function:
            evo = evolved_operator_ansatz(op, evolution=evolution)
        else:
            evo = EvolvedOperatorAnsatz(op, evolution=evolution)

        parameters = evo.parameters

        reference = QuantumCircuit(3)
        reference.append(HamiltonianGate(matrix, parameters[0]), [0, 1, 2])

        if not use_function:
            evo = evo.decompose().decompose()

        self.assertEqual(evo, reference)

    def test_changing_operators(self):
        """Test rebuilding after the operators changed."""
        ops = [Pauli("X"), Pauli("Y"), Pauli("Z")]
        evo = EvolvedOperatorAnsatz(ops)
        evo.operators = [Pauli("X"), Pauli("Y")]
        parameters = evo.parameters

        reference = QuantumCircuit(1)
        reference.rx(2 * parameters[0], 0)
        reference.ry(2 * parameters[1], 0)

        self.assertEqual(evo.decompose(reps=2), reference)

    @data(True, False)
    def test_invalid_reps(self, use_function):
        """Test setting an invalid number of reps."""
        with self.assertRaises(ValueError):
            if use_function:
                _ = evolved_operator_ansatz(Pauli("X"), reps=-1)
            else:
                _ = EvolvedOperatorAnsatz(Pauli("X"), reps=-1)

    def test_insert_barriers_circuit(self):
        """Test using insert_barriers."""
        evo = EvolvedOperatorAnsatz(Pauli("Z"), reps=4, insert_barriers=True)
        ref = QuantumCircuit(1)
        for parameter in evo.parameters:
            ref.rz(2.0 * parameter, 0)
            ref.barrier()

        self.assertEqual(evo.decompose(reps=2), ref)

    def test_insert_barriers(self):
        """Test using insert_barriers."""
        evo = evolved_operator_ansatz(Pauli("Z"), reps=4, insert_barriers=True)
        ref = QuantumCircuit(1)
        for i, parameter in enumerate(evo.parameters):
            ref.rz(2.0 * parameter, 0)
            if i < evo.num_parameters - 1:
                ref.barrier()

        self.assertEqual(evo, ref)

    def test_empty_build_fails(self):
        """Test setting no operators to evolve raises the appropriate error."""
        evo = EvolvedOperatorAnsatz()
        with self.assertRaises(ValueError):
            _ = evo.draw()

    @data(True, False)
    def test_empty_operator_list(self, use_function):
        """Test setting an empty list of operators to be equal to an empty circuit."""
        if use_function:
            evo = evolved_operator_ansatz([])
        else:
            evo = EvolvedOperatorAnsatz([])

        self.assertEqual(evo, QuantumCircuit())

    @data(True, False)
    def test_matrix_operator(self, use_function):
        """Test passing a quantum_info.Operator uses the HamiltonianGate."""
        unitary = Operator([[0, 1], [1, 0]])

        if use_function:
            evo = evolved_operator_ansatz(unitary, reps=3)
        else:
            evo = EvolvedOperatorAnsatz(unitary, reps=3).decompose()

        self.assertEqual(evo.count_ops()["hamiltonian"], 3)

    def test_flattened(self):
        """Test flatten option is actually flattened."""
        num_qubits = 3
        ops = [Pauli("Z" * num_qubits), Pauli("Y" * num_qubits), Pauli("X" * num_qubits)]
        evo = EvolvedOperatorAnsatz(ops, reps=3, flatten=True)
        self.assertNotIn("hamiltonian", evo.count_ops())
        self.assertNotIn("EvolvedOps", evo.count_ops())
        self.assertNotIn("PauliEvolution", evo.count_ops())

    def test_flattening(self):
        """Test ``flatten`` on the function."""
        operators = [Operator.from_label("X"), SparsePauliOp(["Z"])]

        with self.subTest(flatten=None):
            evo = evolved_operator_ansatz(operators, flatten=None)
            ops = evo.count_ops()
            self.assertIn("hamiltonian", ops)
            self.assertNotIn("PauliEvolution", ops)

        with self.subTest(flatten=True):
            # check we get a warning when trying to flatten a HamiltonianGate,
            # which has an unbound param and cannot be flattened
            with self.assertWarnsRegex(UserWarning, "Cannot flatten"):
                evo = evolved_operator_ansatz(operators, flatten=True)

            ops = evo.count_ops()
            self.assertIn("hamiltonian", ops)
            self.assertNotIn("PauliEvolution", ops)

        with self.subTest(flatten=False):
            evo = evolved_operator_ansatz(operators, flatten=False)
            ops = evo.count_ops()
            self.assertIn("hamiltonian", ops)
            self.assertIn("PauliEvolution", ops)


class TestHamiltonianVariationalAnsatz(QiskitTestCase):
    """Test the hamiltonian_variational_ansatz function.

    This is essentially already tested by the evolved_operator_ansatz, we just need
    to test the additional commuting functionality.
    """

    def test_detect_commutation(self):
        """Test the operator is split into commuting terms."""
        hamiltonian = SparsePauliOp(["XII", "ZZI", "IXI", "IZZ", "IIX"])
        circuit = hamiltonian_variational_ansatz(hamiltonian)

        # this Hamiltonian should be split into 2 commuting groups, hence we get 2 parameters
        self.assertEqual(2, circuit.num_parameters)

    def test_evolution_with_identity(self):
        """Test a Hamiltonian containing an identity term.

        Regression test of #13644.
        """
        hamiltonian = SparsePauliOp(["III", "IZZ", "IXI"])
        ansatz = hamiltonian_variational_ansatz(hamiltonian, reps=1)
        bound = ansatz.assign_parameters([1, 1])  # we have two non-commuting groups, hence 2 params

        expected = QuantumCircuit(3, global_phase=-1)
        expected.rzz(2, 0, 1)
        expected.rx(2, 1)

        self.assertEqual(expected, bound)


def evolve(pauli_string, time):
    """Get the reference evolution circuit for a single Pauli string."""

    num_qubits = len(pauli_string)
    forward = QuantumCircuit(num_qubits)
    for i, pauli in enumerate(pauli_string):
        if pauli == "x":
            forward.h(i)
        elif pauli == "y":
            forward.sx(i)

    for i in range(1, num_qubits):
        forward.cx(num_qubits - i, num_qubits - i - 1)

    circuit = QuantumCircuit(num_qubits)
    circuit.compose(forward, inplace=True)
    circuit.rz(2 * time, 0)
    circuit.compose(forward.inverse(), inplace=True)

    return circuit
